"""Uses ctypes and libespeak-ng to get IPA phonemes from text"""
import ctypes
import re
from enum import Enum
from pathlib import Path
from typing import Any, Collection, List, Optional

_DIR = Path(__file__).parent
__version__ = (_DIR / "VERSION").read_text().strip()

# -----------------------------------------------------------------------------


class StreamType(Enum):
    """Type of stream used to record phonemes from eSpeak."""

    MEMORY = "memory"
    NONE = "none"


class Phonemizer:
    """
    Use ctypes and libespeak-ng to get IPA phonemes from text.
    Not thread safe.

    Tries to use libespeak-ng.so or libespeak-ng.so.1
    """

    SEEK_SET = 0

    EE_OK = 0

    AUDIO_OUTPUT_SYNCHRONOUS = 0x02
    espeakPHONEMES_IPA = 0x02
    espeakCHARS_AUTO = 0
    espeakSSML = 0x10
    espeakPHONEMES = 0x100

    LANG_SWITCH_FLAG = re.compile(r"\([^)]*\)")

    DEFAULT_CLAUSE_BREAKERS = {",", ";", ":", ".", "!", "?"}

    STRESS_PATTERN = re.compile(r"[ˈˌ]")

    def __init__(
        self,
        default_voice: Optional[str] = None,
        clause_breakers: Optional[Collection[str]] = None,
        stream_type: StreamType = StreamType.MEMORY,
    ):
        self.current_voice: Optional[str] = None
        self.default_voice = default_voice
        self.clause_breakers = clause_breakers or Phonemizer.DEFAULT_CLAUSE_BREAKERS

        self.stream_type = stream_type
        self.libc: Any = None
        self.lib_espeak: Any = None

    def phonemize(
        self,
        text: str,
        voice: Optional[str] = None,
        keep_clause_breakers: bool = False,
        phoneme_separator: Optional[str] = None,
        word_separator: str = " ",
        punctuation_separator: str = "",
        keep_language_flags: bool = False,
        no_stress: bool = False,
        ssml: bool = False,
    ) -> str:
        """
        Return IPA string for text.
        Not thread safe.

        Args:
            text: Text to phonemize
            voice: optional voice (uses self.default_voice if None)
            keep_clause_breakers: True if punctuation symbols should be kept
            phoneme_separator: Separator character between phonemes
            word_separator: Separator string between words (default: space)
            punctuation_separator: Separator string between before punctuation (keep_clause_breakers=True)
            keep_language_flags: True if language switching flags should be kept
            no_stress: True if stress characters should be removed

        Returns:
            ipa - string of IPA phonemes
        """
        if ssml and (self.stream_type == StreamType.NONE):
            raise ValueError("Cannot use SSML without stream")

        self._maybe_init()

        voice = voice or self.default_voice

        if (voice is not None) and (voice != self.current_voice):
            self.current_voice = voice
            voice_bytes = voice.encode("utf-8")
            result = self.lib_espeak.espeak_SetVoiceByName(voice_bytes)
            assert result == Phonemizer.EE_OK, f"Failed to set voice to {voice}"

        missing_breakers = []
        if keep_clause_breakers and self.clause_breakers:
            missing_breakers = [c for c in text if c in self.clause_breakers]

        phoneme_flags = Phonemizer.espeakPHONEMES_IPA
        if phoneme_separator:
            phoneme_flags = phoneme_flags | (ord(phoneme_separator) << 8)

        if self.stream_type == StreamType.MEMORY:
            phoneme_lines = self._phonemize_mem_stream(text, phoneme_separator, ssml)
        elif self.stream_type == StreamType.NONE:
            phoneme_lines = self._phonemize_no_stream(text, phoneme_separator)
        else:
            raise ValueError("Unknown stream type")

        if not keep_language_flags:
            # Remove language switching flags, e.g. (en)
            phoneme_lines = [
                Phonemizer.LANG_SWITCH_FLAG.sub("", line) for line in phoneme_lines
            ]

        if word_separator != " ":
            # Split/re-join words
            for line_idx in range(len(phoneme_lines)):
                phoneme_lines[line_idx] = word_separator.join(
                    phoneme_lines[line_idx].split()
                )

        # Re-insert clause breakers
        if missing_breakers:
            # pylint: disable=consider-using-enumerate
            for line_idx in range(len(phoneme_lines)):
                if line_idx < len(missing_breakers):
                    phoneme_lines[line_idx] += (
                        punctuation_separator + missing_breakers[line_idx]
                    )

        phonemes_str = word_separator.join(line.strip() for line in phoneme_lines)

        if no_stress:
            # Remove primary/secondary stress markers
            phonemes_str = Phonemizer.STRESS_PATTERN.sub("", phonemes_str)

        # Clean up multiple phoneme separators
        if phoneme_separator:
            phonemes_str = re.sub(
                "[" + re.escape(phoneme_separator) + "]+",
                phoneme_separator,
                phonemes_str,
            )

        return phonemes_str

    def _phonemize_mem_stream(
        self, text: str, phoneme_separator: Optional[str], ssml: bool
    ) -> List[str]:
        # Create in-memory file for phoneme trace.
        phonemes_buffer = ctypes.c_char_p()
        phonemes_size = ctypes.c_size_t()
        phonemes_file = self.libc.open_memstream(
            ctypes.byref(phonemes_buffer), ctypes.byref(phonemes_size)
        )

        try:
            phoneme_flags = Phonemizer.espeakPHONEMES_IPA
            if phoneme_separator:
                phoneme_flags = phoneme_flags | (ord(phoneme_separator) << 8)

            self.lib_espeak.espeak_SetPhonemeTrace(phoneme_flags, phonemes_file)

            text_bytes = text.encode("utf-8")

            synth_flags = Phonemizer.espeakCHARS_AUTO | Phonemizer.espeakPHONEMES
            if ssml:
                synth_flags |= Phonemizer.espeakSSML

            self.lib_espeak.espeak_Synth(
                text_bytes,
                0,  # buflength (unused in AUDIO_OUTPUT_SYNCHRONOUS mode)
                0,  # position
                0,  # position_type
                0,  # end_position (no end position)
                synth_flags,
                None,  # unique_speaker,
                None,  # user_data,
            )
            self.libc.fflush(phonemes_file)

            return ctypes.string_at(phonemes_buffer).decode().splitlines()
        finally:
            self.libc.fclose(phonemes_file)

    def _phonemize_no_stream(
        self, text: str, phoneme_separator: Optional[str]
    ) -> List[str]:
        phoneme_flags = Phonemizer.espeakPHONEMES_IPA
        if phoneme_separator:
            phoneme_flags = phoneme_flags | (ord(phoneme_separator) << 8)

        text_bytes = text.encode("utf-8")
        text_pointer = ctypes.c_char_p(text_bytes)
        text_flags = Phonemizer.espeakCHARS_AUTO
        phoneme_lines = []
        fcn_ttp = self.lib_espeak.espeak_TextToPhonemes
        fcn_ttp.restype = ctypes.c_char_p
        while text_pointer:
            clause_phonemes = fcn_ttp(
                ctypes.pointer(text_pointer), text_flags, phoneme_flags
            )
            if isinstance(clause_phonemes, bytes):
                clause_phonemes_str = (
                    clause_phonemes.decode()
                )  # pylint: disable=no-member
                phoneme_lines.append(clause_phonemes_str)

        return phoneme_lines

    def _maybe_init(self):
        if self.lib_espeak:
            # Already initialized
            return

        try:
            self.lib_espeak = ctypes.cdll.LoadLibrary("libespeak-ng.so")
        except OSError:
            # Try .so.1
            self.lib_espeak = ctypes.cdll.LoadLibrary("libespeak-ng.so.1")

        sample_rate = self.lib_espeak.espeak_Initialize(
            Phonemizer.AUDIO_OUTPUT_SYNCHRONOUS, 0, None, 0
        )
        assert sample_rate > 0, "Failed to initialize libespeak-ng"

        if self.stream_type == StreamType.MEMORY:
            # Initialize libc for memory stream
            self.libc = ctypes.cdll.LoadLibrary("libc.so.6")
            self.libc.open_memstream.restype = ctypes.POINTER(ctypes.c_char)



"""Command-line interface to espeak_phonemizer"""
import argparse
import csv
import logging
import os
import sys

from . import Phonemizer

_LOGGER = logging.getLogger("espeak_phonemizer")

# -----------------------------------------------------------------------------


def main():
    parser = argparse.ArgumentParser(prog="espeak_phonemizer")
    parser.add_argument("-v", "--voice", help="eSpeak voice to use")
    parser.add_argument(
        "-p", "--phoneme-separator", help="Separator character between phonemes"
    )
    parser.add_argument(
        "-w",
        "--word-separator",
        help="Separator string between words (cannot be used if --phoneme-separator is whitespace)",
    )
    parser.add_argument(
        "--keep-punctuation",
        action="store_true",
        help="Keep clause-breaking punctuation characters (,;:.!?)",
    )
    parser.add_argument(
        "--keep-language-flags",
        action="store_true",
        help="Keep language-switching flags",
    )
    parser.add_argument(
        "--print-input", action="store_true", help="Print input text before phonemes"
    )
    parser.add_argument(
        "--output-separator",
        default=" ",
        help="Separator string between input text and phonemes",
    )
    parser.add_argument(
        "--no-stress",
        action="store_true",
        help="Remove primary/secondary stress markers",
    )
    parser.add_argument(
        "--csv",
        action="store_true",
        help="Input and output is CSV. Phonemes are added as a final column",
    )
    parser.add_argument(
        "--csv-delimiter", default="|", help="Delimiter in CSV input and output"
    )
    parser.add_argument("--version", action="store_true", help="Print version and exit")
    parser.add_argument(
        "--debug", action="store_true", help="Print DEBUG messages to the console"
    )
    args = parser.parse_args()

    if args.debug:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)

    # -------------------------------------------------------------------------

    if args.version:
        # Print version and exit
        from . import __version__

        print(__version__)
        sys.exit(0)

    # -------------------------------------------------------------------------

    assert args.voice, "Missing -v/--voice"

    if args.word_separator:
        assert (
            args.phoneme_separator.strip()
        ), "Word separator cannot be used if phoneme separator is whitespace"

    phonemizer = Phonemizer(default_voice=args.voice)

    # CSV input/output
    if args.csv:
        csv_writer = csv.writer(sys.stdout, delimiter=args.csv_delimiter)
        reader = csv.reader(sys.stdin, delimiter=args.csv_delimiter)
    else:
        csv_writer = None
        reader = sys.stdin

    if os.isatty(sys.stdin.fileno()):
        print("Reading text from stdin...", file=sys.stderr)

    for line in reader:
        if args.csv:
            text = line[-1]
        else:
            text = line.strip()
            if not text:
                continue

        text_phonemes = phonemizer.phonemize(
            text,
            keep_clause_breakers=args.keep_punctuation,
            phoneme_separator=args.phoneme_separator,
            keep_language_flags=args.keep_language_flags,
            no_stress=args.no_stress,
            punctuation_separator=args.phoneme_separator or "",
        )

        if args.word_separator:
            text_phonemes = args.word_separator.join(text_phonemes.split())

        if args.csv:
            assert csv_writer is not None
            csv_writer.writerow((*line, text_phonemes))
        else:
            if args.print_input:
                print(text, text_phonemes, sep=args.output_separator)
            else:
                print(text_phonemes)


# -----------------------------------------------------------------------------

if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""Tests for Phonemizer class"""
import unittest

from espeak_phonemizer import Phonemizer, StreamType

_STREAM_TYPES = [StreamType.MEMORY, StreamType.NONE]


class PhonemizerTestCase(unittest.TestCase):
    """Test cases for Phonemizer"""

    def test_en(self):
        """Test basic English"""
        for stream_type in _STREAM_TYPES:
            phonemizer = Phonemizer(default_voice="en-us", stream_type=stream_type)
            phonemes = phonemizer.phonemize("test")
            self.assertEqual(phonemes, "tˈɛst")

    def test_no_stress(self):
        """Test stress removal"""
        for stream_type in _STREAM_TYPES:
            phonemizer = Phonemizer(default_voice="en-us", stream_type=stream_type)
            phonemes = phonemizer.phonemize("test", no_stress=True)
            self.assertEqual(phonemes, "tɛst")

    def test_phoneme_separator(self):
        """Test with phoneme separator"""
        for stream_type in _STREAM_TYPES:
            phonemizer = Phonemizer(default_voice="en-us", stream_type=stream_type)
            phonemes = phonemizer.phonemize("test", phoneme_separator="_")
            self.assertEqual(phonemes, "t_ˈɛ_s_t")

    def test_word_phoneme_separators(self):
        """Test with word and phoneme separators"""
        for stream_type in _STREAM_TYPES:
            phonemizer = Phonemizer(default_voice="en-us", stream_type=stream_type)
            phonemes = phonemizer.phonemize(
                "test 1", phoneme_separator="_", word_separator="#"
            )
            self.assertEqual(phonemes, "t_ˈɛ_s_t#w_ˈʌ_n")

    def test_keep_clause_breakers(self):
        """Test keeping punctuation characters that break apart clauses"""
        for stream_type in _STREAM_TYPES:
            phonemizer = Phonemizer(default_voice="en-us", stream_type=stream_type)
            phonemes = phonemizer.phonemize("test: 1, 2, 3!", keep_clause_breakers=True)
            self.assertEqual(phonemes, "tˈɛst: wˈʌn, tˈuː, θɹˈiː!")

    def test_keep_language_flags(self):
        """Test keeping language-switching flags"""
        for stream_type in _STREAM_TYPES:
            phonemizer = Phonemizer(default_voice="fr", stream_type=stream_type)

            # Without language flags
            phonemes = phonemizer.phonemize("library")
            self.assertEqual(phonemes, "lˈaɪbɹəɹi")

            # With language flags
            phonemes = phonemizer.phonemize("library", keep_language_flags=True)
            self.assertEqual(phonemes, "(en)lˈaɪbɹəɹi(fr)")

    def test_ssml(self):
        """Test SSML"""
        phonemizer = Phonemizer(default_voice="en-us", stream_type=StreamType.MEMORY)
        phonemes = phonemizer.phonemize(
            '<speak><s><say-as interpret-as="characters">12</say-as></s></speak>',
            phoneme_separator="_",
            ssml=True,
        )

        # "one two" instead of "twelve"
        phoneme_list = list(filter(None, (p.strip() for p in phonemes.split("_"))))
        self.assertEqual(phoneme_list, ["w", "ˌʌ", "n", "t", "ˈuː"])



"""Setup file for espeak_phonemizer"""
import os
from pathlib import Path

import setuptools

this_dir = Path(__file__).parent
module_dir = this_dir / "espeak_phonemizer"

# -----------------------------------------------------------------------------

# Load README in as long description
long_description: str = ""
readme_path = this_dir / "README.md"
if readme_path.is_file():
    long_description = readme_path.read_text()

version_path = module_dir / "VERSION"
with open(version_path, "r") as version_file:
    version = version_file.read().strip()

# -----------------------------------------------------------------------------

setuptools.setup(
    name="espeak_phonemizer",
    version=version,
    description="Lightweight International Phonetic Alphabet (IPA) phonemizer that uses libespeak-ng",
    author="Michael Hansen",
    author_email="mike@rhasspy.org",
    url="https://github.com/rhasspy/espeak-phonemizer",
    packages=setuptools.find_packages(),
    package_data={"espeak_phonemizer": ["VERSION", "py.typed"]},
    entry_points={
        "console_scripts": [
            "espeak-phonemizer = espeak_phonemizer.__main__:main",
        ]
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
    ],
    long_description=long_description,
    long_description_content_type="text/markdown",
)
